import os
import glob
import subprocess
import urllib.request
import urllib.error
import yaml
from reseasdk.helpers import load_yaml
from reseasdk.validators import \
    validate_packages_yml, validate_package_yml, \
    validate_build_yml, validate_global_yml, ValidationError
from reseasdk.helpers import error


def get_path_from_packages_yml(package):
    yml = load_yaml('packages.yml', validator=validate_packages_yml)

    try:
        path = os.path.join(os.getcwd(), yml['packages'][package]['path'])
    except KeyError:
        return None

    return path


package_uris = {} # cache
PACKAGE_DB_URL = "https://raw.githubusercontent.com/resea/packages/master/{}.yml"
def _download_package(package):
    if package_uris.get(package) is None:
        # get package's URI from PACKAGE_DB_URI
        try:
            r = urllib.request.urlopen(PACKAGE_DB_URL.format(package))
            uri = yaml.safe_load(r)['uri']
        except urllib.error.HTTPError as e:
            error("failed to download package '{}': {}".format(package, str(e)))
        except KeyError:
            error("cannot download package '{}': uri not found in the package "
                  "database".format(package))

        # download the package
        scheme, repo = uri.split(':', 2)
        path = scheme + "-" + uri.replace('/', '-').replace(':', '-')
        if scheme == 'github':
            subprocess.check_output(['git', 'clone',
                                     'https://github.com/' + repo,
                                     path])
        else:
            error("cannot download package '{}': unknown uri scheme '{}'".format(
                  package, scheme))


def search_vendor_dir(package):
    # search .vendor directory
    cwd = os.getcwd()
    for d in glob.iglob('packages/.vendor/*'):
        os.chdir(d)
        path = get_path_from_packages_yml(package)
        os.chdir(cwd)
        if path:
            return path
    return None


def download_package(package):
    path = search_vendor_dir(package)
    if path:
        return path
    else:
        # download the package
        os.makedirs('packages/.vendor', exist_ok=True)
        os.chdir('packages/.vendor')
        _download_package(package)
        os.chdir('../..')
        path = search_vendor_dir(package)
        if path:
            return path
        else:
            error("failed to search for package '{}'".format(package))


current_package_name = None # cache
def get_package_dir(package):
    global current_package_name

    cwd = os.getcwd()

    # the current package
    if current_package_name is None:
        current_package_name = load_yaml('package.yml',
                                         validator=validate_package_yml).get('name')
    if current_package_name == package:
        return os.getcwd()

    # look for packages.yml
    while True:
        os.chdir("..")
        if os.getcwd() == "/":
            break
        if os.path.exists("packages.yml"):
            path = get_path_from_packages_yml(package)
            if path:
                os.chdir(cwd)
                return path
            # if not found in the packages.yml, look for its ancestor dirs

    # not found, download it
    os.chdir(cwd)
    return download_package(package)


def get_package(package):
    """ Prepare a package in 'packages' directory """

    if not os.path.exists('packages/' + package):
        os.makedirs('packages', exist_ok=True)
        path = get_package_dir(package)
        os.chdir('packages')
        os.symlink(path, package)
        os.chdir('..')


global_ymls = {}  # cache
def load_global_yml(package):
    if package in global_ymls:
        return global_ymls[package]

    get_package(package)
    try:
        yml = load_yaml('packages/{}/global.yml'.format(package),
                        validator=validate_global_yml)
    except FileNotFoundError:
        yml = {}
    global_ymls[package] = yml
    return yml


build_ymls = {}  # cache
def load_build_yml(package):
    if package in build_ymls:
        return build_ymls[package]

    get_package(package)
    try:
        yml = load_yaml('packages/{}/build.yml'.format(package),
                        validator=validate_build_yml)
    except FileNotFoundError:
        yml = {}
    build_ymls[package] = yml
    return yml


package_ymls = {}  # cache
def load_package_yml(package):
    if package in package_ymls:
        return package_ymls[package]

    get_package(package)
    yml = load_yaml('packages/{}/package.yml'.format(package),
                    validator=validate_package_yml)
    package_ymls[package] = yml
    return yml


def resolve_package_dependency(_packages, all_in_one, interfaces_only=False):
    """ Returns list of required packages. """
    packages = []
    follows = ['library']
    if all_in_one:
        follows.append('application')
    for package in _packages:
        yml = load_package_yml(package)
        packages += yml['implements']
        packages += yml['uses']
        for require in yml['requires'] + yml['implements'] + yml['uses']:
            if interfaces_only or \
               load_package_yml(require)['category'] in follows:
                packages += resolve_package_dependency([require], all_in_one,
                                                       interfaces_only=interfaces_only)
            packages.append(require)
        packages.append(package)
    return list(set(packages))


def filter_package_by_category(categories, packages):
    matched = []
    for package in packages:
        if load_package_yml(package)['category'] in categories:
            matched.append(package)
    return matched



import os
import sys
import argparse
import subprocess
from reseasdk.package import load_package_yml
from reseasdk.validators import validate_package_yml
from reseasdk.helpers import load_yaml, error


SHORT_HELP = "Generate boilterplate files"
LONG_HELP = """
Usage: reseasdk scaffold
"""


def scaffold(args):
    if not os.path.exists('package.yml'):
        error("'package.yml' not found (are you in a package directory?)")
    yml = load_yaml('package.yml', validator=validate_package_yml)

    lang = yml.get("lang")
    if lang is None:
        error("lang is not speicified in package.yml")

    bin = load_package_yml(lang)["scaffold"]
    subprocess.Popen([bin, ".", "package.yml"]).wait()


def main(args_):
    parser = argparse.ArgumentParser(prog='reseasdk scaffold',
                                     description='Generate boilerplate files')
    args = parser.parse_args(args_)

    scaffold(parser.parse_args(args_))



import shutil


SHORT_HELP = "remove generated files during builds"
LONG_HELP = """
Usage: reseasdk clean
"""


def clean():
    try:
        shutil.rmtree('build')
    except FileNotFoundError:
        pass


def main(args):
    clean()



import argparse
import os
import re
import sys
import jinja2
import reseasdk
from reseasdk.helpers import error, generating, exec_cmd


SHORT_HELP = "generate a porject boilterplate files"
LONG_HELP = """
Usage: reseasdk new package_name
"""


README_TEMPLATE = """\
# {{ package_name }}

"""

GITIGNORE_TEMPLATE = """\
/build
/packages
"""

PACKAGE_YML_TEMPLATE = """\
name: {{ package_name }}
category:    # application, library or interface
# id:
license:     # MIT, BSD 2-clause, BSD 3-clause, Public Domain, GPLv2
summary:     # brief explanation
description: # longer explanation

warning:     # some important things to note
info:        # what we should know

author: {{ author }}
email: {{ email }}
homepage:    # e.g. http://example.com/foo/bar

# Prefered language
# lang: c

# Required packages
requires: []

# Interfaces which this package uses
uses: []

# Interfaces which this package implements
implements: []


#
#  Type definitions
#
type:


#
#  Interface definitions
#
interface:
"""

FILES = [
    # (filepath, template)
    ('README.md', README_TEMPLATE),
    ('.gitignore', GITIGNORE_TEMPLATE),
    ('package.yml', PACKAGE_YML_TEMPLATE),
    ('src/.gitkeep', ''),
    ('Documentation/.gitkeep', '')
]

PACKAGE_NAME_REGEX = '[a-z_][a-z0-9_]*'


def main(args_):
    parser = argparse.ArgumentParser(prog='reseasdk new',
                                     description='create a new package')
    parser.add_argument('--git', action='store_true',
                        help='git init in the created directory')
    parser.add_argument('dir', help='directory')
    args = parser.parse_args(args_)
    package_name = os.path.basename(args.dir)

    if re.match(PACKAGE_NAME_REGEX, package_name) is None:
        error('The package name must be lowercase_with_underscore (regex: {})'
              .format(PACKAGE_NAME_REGEX))

    #
    #  generate the package directory
    #
    generating('MKDIR', package_name)
    try:
        os.mkdir(package_name)
    except FileExistsError:
        error("Directory already exists: '{}'".format(package_name))

    #
    #  Variables used in templates
    #
    author = exec_cmd('git config --get user.name',
                      ignore_failure=True).strip()
    email = exec_cmd('git config --get user.email',
                     ignore_failure=True).strip()

    #
    #  generate directories and files
    #
    for path, tmpl in FILES:
        f = '{}/{}'.format(package_name, path)
        dir = os.path.dirname(f)

        if not os.path.exists(dir):
            generating('MKDIR', dir)
            os.mkdir(dir)

        generating('GEN', f)
        open(f, 'w').write(jinja2.Template(tmpl).render(**locals()))

    # intialize as a new Git repository
    if args.git:
        exec_cmd('git init .', cwd=package_name)



from reseasdk.version import __version__
from reseasdk.main import main

import reseasdk.defines
import reseasdk.helpers
import reseasdk.validators
import reseasdk.package

command_list = [
    'build',
    'test',
    'clean',
    'log',
    'new',
    'scaffold',
    'version'
]



import argparse
import hashlib
import os
import sys
import time
import threading
import watchdog
import watchdog.events
import watchdog.observers
from reseasdk.run import run_emulator
from reseasdk.helpers import info, notice, error, \
    dict_to_strdict
from reseasdk.commands.build import build


SHORT_HELP = "build and run"
LONG_HELP = """
Usage: reseasdk run
"""


def run(args):
    cs = build(args)
    defaults = cs['defaults']

    info('ReseaSDK: build finished successfully, starting the executable...')
    env = dict_to_strdict(defaults)
    cmd = [defaults['HAL_RUN'], defaults['BUILD_DIR'] + '/application']
    run_emulator(cmd, save_log='build/{}/boot.log'.format(args['target']),
                 env=env)


def main(args_):
    parser = argparse.ArgumentParser(prog='reseasdk run',
                                     description='run an executable')
    parser.add_argument('-r', action='store_true',
                        help='rebuild the executable')
    parser.add_argument('--target', default='build', help='the build target')
    parser.add_argument('--dump', action='store_true',
                        help='dump generated files for debugging')
    parser.add_argument('--no-prettify', action='store_true',
                        help="don't prettify output")
    parser.add_argument('--all-in-one', action='store_true',
                        help='embed all required applications so that'
                        'discovery() always succeed')
    parser.add_argument('--single-app', action='store_true',
                        help='no threading')
    parser.add_argument('defaults', nargs='*',
                        help='default variables (FOO=bar)')

    args = vars(parser.parse_args(args_))
    run(args)
